
#ifdef __KERNEL__
#include <linux/module.h>
#include <linux/types.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/ip.h>     /* Needed here. Otherwise the linksys target fails! */
#include <linux/time.h>
#else
#include <stdlib.h>
#include <stdio.h>
#endif

#include "pla_conn.h"

static spinlock_t conn_lock;
static struct list_head conn_list;

static struct pla_conn* conn_find(char *key)
{
	struct list_head *cursor;

	list_for_each(cursor, &conn_list) {
		struct pla_conn *this_entry = (struct pla_conn*) cursor;

		if (memcmp(this_entry->key, key, sizeof(this_entry->key) ) == 0) {
			return this_entry;
		}
	}

	return NULL;
}

/*
 * Find entry from db. If not found, add a new entry.
 */
int pla_conn_find_add(char *key, uint64_t seqnum, uint32_t c_time)
{
	int advance, nulled_entries;
	int i;

	int retval;
	struct pla_conn *entry;
	int window_high;   /* Highest sequence number that can be put in rep_win */

	struct timeval tv;

	/* Check timestamp */
	do_gettimeofday(&tv);
	if (c_time > tv.tv_sec + PLA_TIMESTAMP_DIFF || c_time < tv.tv_sec - PLA_TIMESTAMP_DIFF) {
		printk(KERN_ALERT "Time difference is %d seconds\n", (int) (tv.tv_sec - c_time) );
		retval = -1;
		goto out_err_no_unlock;
	}

	spin_lock_bh(&conn_lock);

	/* Find entry. If not available, create it. */
	if ( !(entry = conn_find(key)) ) {
		entry = kmalloc(sizeof(struct pla_conn), GFP_ATOMIC);
		if (!entry) {
			retval = -ENOMEM;
			goto out_err;
		}
		memcpy(entry->key, key, sizeof(entry->key));
		memset(entry->rep_win, 0, sizeof(entry->rep_win));
		entry->window_low = 0;
		entry->low_index = 0;

		list_add(&entry->list_head, &conn_list);
	}

	/* Seqnum is below window: drop */
	if (seqnum < entry->window_low) {
		retval = -1;
		goto out;
	}
	
	window_high = entry->window_low + PLA_WINDOW_SIZE - 1;

	/* Seqnum within window: */
	if (seqnum <= window_high) {
		int entry_index = (entry->low_index + (seqnum-entry->window_low)) % PLA_WINDOW_SIZE;

		if (entry->rep_win[entry_index] == 0) {
			entry->rep_win[entry_index] = 1;
			retval = 0;    /* Accept new packet */
		} else {
			retval = -1;   /* Drop duplicate */
		}
		goto out;
	}

	/* Seqnum is above window: Advance window. */
	advance = seqnum - window_high;

	/* Zero the new area */
	nulled_entries = min(PLA_WINDOW_SIZE, advance);
	for (i=0; i<nulled_entries; i++)
		entry->rep_win[(entry->low_index + i) % PLA_WINDOW_SIZE] = 0;
	
	entry->low_index = (entry->low_index + advance) % PLA_WINDOW_SIZE;
	entry->window_low += advance;
	entry->rep_win[(entry->low_index + (seqnum-entry->window_low)) % PLA_WINDOW_SIZE] = 1;

	retval = 0;

out_err:
out:
	spin_unlock_bh(&conn_lock);
out_err_no_unlock:
	return retval;
};

void pla_conn_flush(void)
{
	struct list_head *cursor;
	struct list_head *next;

	spin_lock_bh(&conn_lock);
	list_for_each_safe(cursor, next, &conn_list) {
		struct pla_conn *this_entry = (struct pla_conn*) cursor;

		list_del(cursor);
		kfree(this_entry);
	}
	spin_unlock_bh(&conn_lock);

	return;
}

int pla_conn_init(void)
{
	spin_lock_init(&conn_lock);
	
	INIT_LIST_HEAD(&conn_list);

	return 0;
}

void pla_conn_uninit(void)
{
	pla_conn_flush();
       
	return;
}

#ifndef __KERNEL__
/* User space test code starts here. */
void print_structure(char *key)
{
	int i;
	struct pla_conn *entry;

	entry = conn_find(key);
	if (!entry) {
		fprintf(stderr, "Entry not found!\n");
		exit(-1);
	}

	fprintf(stderr, "%d: ", entry->window_low);
	for (i=entry->low_index; i<entry->low_index+PLA_WINDOW_SIZE; i++)
		if (entry->rep_win[i % PLA_WINDOW_SIZE])
			fprintf(stderr, "%d ", entry->window_low + i - entry->low_index);

	fprintf(stderr, "\n");

	return;
}

int find_add(char *key, int seqnum)
{
	if (pla_conn_find_add(key, seqnum, 0) < 0) {
		fprintf(stderr, "Failed add. Seqnum: %d\n", seqnum);
	} else {
		fprintf(stderr, "Add successful\n");
	}

	return 0;
}

int main(void)
{
	static char key1[PLA_HASH_DIGEST_LENGTH] = {0x12, 0x34, 0x56};

	pla_conn_init();

	find_add(key1, 0);
	find_add(key1, 1);
	find_add(key1, 1);
	find_add(key1, 5);
	find_add(key1, 10);
	find_add(key1, 11);
	find_add(key1, 12);
	find_add(key1, 15); print_structure(key1);

	find_add(key1, 16); print_structure(key1);
	find_add(key1, 20); print_structure(key1);
	find_add(key1, 30); print_structure(key1);
	find_add(key1, 50); print_structure(key1);
	find_add(key1, 54); print_structure(key1);
	find_add(key1, 52); print_structure(key1);

	find_add(key1, 38); print_structure(key1);
	find_add(key1, 39); print_structure(key1);

	pla_conn_uninit();
	return 0;
}
#endif
