#include <linux/netlink.h>
#include <linux/wait.h>
#include <linux/netdevice.h>

#include "pla_policy.h"
#include "pla_netlink.h"

struct pla_policy 
{
	struct list_head list;  /* Must be first in struct */
	struct rcu_head rcu_head;

	int in_if;
	int out_if;
	int policy_rule;
};

LIST_HEAD(policy);
spinlock_t policy_write_lock;
static atomic_t uncleaned_policies = ATOMIC_INIT(0);  /* Number uncleaned entries in policy list */

int pla_policy_add(int in_if, int out_if, int policy_rule)
{
	int err = 0;
	struct pla_policy *p_entry;

	if (policy_rule < 0 || policy_rule > PLA_FORWARD_POLICY_MAX) {
		err = -EINVAL;
		goto out;
	}

	p_entry = kmalloc(sizeof(struct pla_policy), GFP_ATOMIC);
	if (!p_entry) {
		err = -ENOMEM;
		goto out;
	}
		
	p_entry->in_if = in_if;
	p_entry->out_if = out_if;
	p_entry->policy_rule = policy_rule;

	spin_lock_bh(&policy_write_lock);
	list_add_rcu( &p_entry->list, &policy);
	atomic_inc(&uncleaned_policies);
	spin_unlock_bh(&policy_write_lock);

out:
	return err;
}

int pla_policy_netlink(struct nlmsghdr *nlh)
{
	int err = 0;
	int type;
	struct pla_netlink_set_interface *if_set;

	type = nlh->nlmsg_type;
	switch(type) {
	case NLMSG_PLA_SET_INTERFACES:
		if_set = NLMSG_DATA(nlh);
		pla_policy_add(if_set->if_in, if_set->if_out, if_set->policy);
		break;

	default:
		break;
	}

	return err;
}

static void pla_policy_free_entry(struct rcu_head *head)
{
	kfree(container_of(head, struct pla_policy, rcu_head));
	atomic_dec(&uncleaned_policies);

	return;
}

int pla_policy_flush_complete(void)
{
	return atomic_read(&uncleaned_policies) == 0 ? 1 : 0;
}

void pla_policy_flush(void)
{
	struct pla_policy *e;
	struct pla_policy *safe_storage;

	list_for_each_entry_safe(e, safe_storage, &policy, list) {
		list_del_rcu(&e->list);
		call_rcu( &e->rcu_head, pla_policy_free_entry);
	}

	return;
}

int pla_forward_policy(struct nf_info *info) 
{
	struct list_head *pos;
	int ret = PLA_FORWARD_POLICY_NONE;

	int dev_in_if = info->indev ? info->indev->ifindex : 0;
	int dev_out_if = info->outdev ? info->outdev->ifindex : 0;

	rcu_read_lock();

	list_for_each_rcu(pos, &policy) {
		struct pla_policy *p = (struct pla_policy*) pos;

		if (dev_in_if == p->in_if && dev_out_if == p->out_if) {
			ret = p->policy_rule;
			goto out;
		}

	}

out:
	rcu_read_unlock();
	return ret;
}

void pla_policy_uninit(void)
{
	wait_queue_head_t wqh;

	pla_policy_flush();
	/* 
	 * Wait until all RCU-callbacks have completed.
	 * If the module exits too early, the callbacks will call nonexisting functions.
	 */
	init_waitqueue_head(&wqh);
	while (!pla_policy_flush_complete())
		wait_event_timeout(wqh, 0, HZ/10);

	return;
}

int pla_policy_init(void)
{
	spin_lock_init(&policy_write_lock);

	return 0;
}

