#include <linux/list.h>
#include <linux/delay.h>
#include <net/protocol.h>

#include <net/ipv6.h>
#include <linux/netlink.h>
#include <linux/in.h>
#include <linux/ip.h>

#include "pla_netlink.h"
#include "pla_crypto.h"
#include "pla_crypto_impl.h"
#include "pla_userspace_crypto.h"

static struct workqueue_struct *wqs;

struct req_queue_entry
{
	struct list_head list;                             /* Must be first in struct */
	int payload_len;                                   /* Length of payload */

	struct userspace_crypto_req req;
};

struct userspace_crypto_state
{
	struct list_head list;                          /* Request list pointers. Must be first in struct. */
	struct timer_list timer;                        /* Request expiry timer  */

	struct pla_crypto_task *task;

	void *state;
	void (*callback) (void *call_state);

	__u32 id_num;                                   /* Id number of the expected packet */
};

/* List of requests waiting to be sent by the workqueue */
static LIST_HEAD(pending_reqs);
static spinlock_t pending_reqs_lock;

/* List of expected replys */
static LIST_HEAD(pending_replys);
static spinlock_t pending_replys_lock;

/* Destionation address for crypto requests */
static uint ucrypto_dst;
static spinlock_t usercrypto_lock;

static DEFINE_PER_CPU(struct socket*, userspace_crypto_socket) = NULL;
static DEFINE_PER_CPU(u32, msg_id_num) = 0;

void userspace_crypto_sendmsg(void *unused)
{
	struct sockaddr_in sin;
	struct msghdr msg;
	struct iovec msgvec;

	struct socket *sk;

	struct req_queue_entry *req_entry;

	/* Send all waiting requests */
	while(1) {
		int payload_len;

		spin_lock_bh(&pending_reqs_lock);

		if(list_empty(&pending_reqs)) {
			spin_unlock_bh(&pending_reqs_lock);
			goto out;
		}

		req_entry = (struct req_queue_entry*) pending_reqs.next;
		list_del(pending_reqs.next);
		spin_unlock_bh(&pending_reqs_lock); /* kernel_sendmsg must be called non-atomic */

		payload_len = sizeof(struct userspace_crypto_req) + req_entry->payload_len;

		msgvec.iov_base = &req_entry->req;
		msgvec.iov_len = payload_len;
		
		memset(&sin, 0, sizeof(struct sockaddr_in));
		sin.sin_family = PF_INET;

		spin_lock_bh(&usercrypto_lock);
		sin.sin_addr.s_addr = htonl(ucrypto_dst);
		spin_unlock_bh(&usercrypto_lock);
		
		msg.msg_name = (struct sockaddr*) &sin;
		msg.msg_namelen = sizeof(struct sockaddr_in);
		msg.msg_iov = &msgvec;
		msg.msg_iovlen = 1;
		msg.msg_control = NULL;
		msg.msg_controllen = 0;
		msg.msg_flags = 0;

		sk = get_cpu_var(userspace_crypto_socket);
		put_cpu_var(userspace_crypto_socket);
		
#if 0
		kernel_sendmsg(sk, &msg, (struct kvec*) &msgvec, 1, payload_len);
#else
		sock_sendmsg(sk, &msg, payload_len);
#endif
		kfree(&req_entry->list);
	}

out:
	return;
}

DECLARE_WORK(send_work, userspace_crypto_sendmsg, NULL);

struct userspace_crypto_state* replys_list_remove_entry(__u32 id_num) 
{
	struct list_head *ptr;

	spin_lock_bh(&pending_replys_lock);
	
	list_for_each(ptr, &pending_replys) {
		struct userspace_crypto_state *state = (struct userspace_crypto_state*) ptr;
		
		if (state->id_num == id_num) {
			list_del(&state->list);
			spin_unlock_bh(&pending_replys_lock);
			return state;
		}
	}

	/* Not found */
	spin_unlock_bh(&pending_replys_lock);
	return NULL;
}

static void userspace_call_callback(struct userspace_crypto_state *crypto_state)
{
	crypto_state->callback(crypto_state->state);

	kfree(crypto_state);

	return;
}

void userspace_crypto_timeout(unsigned long id_num)
{
	struct userspace_crypto_state *crypto_state;

	crypto_state = replys_list_remove_entry(id_num);
	if (!crypto_state)
		goto out;                          /* Reply packet came first */

	crypto_state->task->type = CRYPTO_TASK_TYPE_RESPONSE;
	crypto_state->task->code = CRYPTO_TASK_CODE_NOT_OK;

	userspace_call_callback(crypto_state);

out:
	return;
}

int userspace_rcv(struct sk_buff *skb)
{
	struct userspace_crypto_state *crypto_state;
	int id_num;
	struct userspace_crypto_reply *reply;

	if (skb->nh.iph->protocol != IPPROTO_USERSPACE_CRYPTO ||
	    !pskb_may_pull(skb, sizeof(struct userspace_crypto_reply)))
                goto out;

	reply = (struct userspace_crypto_reply*) skb->h.raw;

	if (reply->msg_type != USERSPACE_TYPE_REPLY)
		goto out;

	id_num = ntohl(reply->id_num);

	crypto_state = replys_list_remove_entry(id_num);
	if (!crypto_state)                          /* Timer already expired. */
		goto out;

	del_timer(&crypto_state->timer);

	crypto_state->task->type = CRYPTO_TASK_TYPE_RESPONSE;
	crypto_state->task->code = reply->task_code;

	memcpy(crypto_state->task->td.raw_data, reply->resultdata, ntohs(reply->resultdata_len));
	crypto_state->task->data_len = ntohs(reply->resultdata_len);

	userspace_call_callback(crypto_state);

out:
	kfree_skb(skb);
	return 0;
}

int userspace_crypto_call(void (*callback) (void *call_state), 
			  void *state, 
			  struct pla_crypto_task *task)
{
	int err = 0;
	struct userspace_crypto_state *tstate;

	struct req_queue_entry *message;
	struct userspace_crypto_req *req;

	u32 id_num;
	u8 cpu;

	cpu = get_cpu(); put_cpu();
	id_num = get_cpu_var(msg_id_num)++; put_cpu_var(msg_id_num);
	id_num = (id_num & 0x0FFFFFFF) | cpu << 24;

	/* Setup call. */
	tstate = kmalloc(sizeof(struct userspace_crypto_state), GFP_ATOMIC);
	if (!tstate) {
		err = -ENOMEM;
		goto out;
	}
	tstate->callback = callback;
	tstate->state = state;
	tstate->id_num = id_num;

	tstate->task = task;

	/* Setup timer */
	init_timer(&tstate->timer);
	tstate->timer.data = id_num;
	tstate->timer.function = userspace_crypto_timeout;
	tstate->timer.expires = jiffies + PLA_USERSPACE_WAIT;
	add_timer(&tstate->timer);

	/* Setup list */
	spin_lock_bh(&pending_replys_lock);
	list_add(&tstate->list, &pending_replys);
	spin_unlock_bh(&pending_replys_lock);

	/* Send message */
	message = kzalloc(sizeof(struct req_queue_entry) + task->data_len, GFP_ATOMIC);
	if (!message) {
		err = -ENOMEM;
		goto out;
	}

	req = &message->req;
	req->msg_type = USERSPACE_TYPE_REQUEST;
	req->id_num = htonl(id_num);

	req->task_type = task->type;
	req->task_code = task->code;
	memcpy(req->task_data, task->td.raw_data, task->data_len);
	message->payload_len = task->data_len;

	spin_lock_bh(&pending_reqs_lock);
	list_add( (struct list_head*) message, &pending_reqs);
	spin_unlock_bh(&pending_reqs_lock);

	queue_work(wqs, &send_work);

out:
	return err;
}

static struct net_protocol userspace_protocol = {
        .handler =      userspace_rcv,
};

int userspace_crypto_init_uninit(int isinit)
{
	struct sock *sk;
	struct inet_sock *inet;
	int err = 0;
	int i = 0;
	int j;

	if (!isinit)
		goto uninit;

	spin_lock_init(&pending_reqs_lock);
	spin_lock_init(&pending_replys_lock);
	spin_lock_init(&usercrypto_lock);

	wqs = create_workqueue("pla_user");
	if (!wqs)
		goto workqueue_failed;

	for_each_cpu(i) {
		err = sock_create_kern(PF_INET, SOCK_RAW, IPPROTO_USERSPACE_CRYPTO, 
				       &per_cpu(userspace_crypto_socket, i) );
		if (err < 0) {
			printk(KERN_ALERT "Userspace crypto: Failed to create RAW socket: %d\n", err);
			goto sock_err;
		}
		
		sk = per_cpu(userspace_crypto_socket, i)->sk;
		sk->sk_allocation = GFP_ATOMIC;
		sk->sk_prot->unhash(sk);
		
		inet = inet_sk(sk);
		inet->hdrincl = 0;
	}

	if (inet_add_protocol(&userspace_protocol, IPPROTO_USERSPACE_CRYPTO) < 0) {
		printk(KERN_ALERT "Failed to add userspace crypto protocol\n");
		goto proto_add_failed;
	}

	return 0;

uninit:

	inet_del_protocol(&userspace_protocol, IPPROTO_USERSPACE_CRYPTO);
proto_add_failed:


sock_err:
	if (isinit) {
		for (j=0; j<i; j++) {
			sock_release(per_cpu(userspace_crypto_socket, j));
			per_cpu(userspace_crypto_socket, j) = NULL;
		}
	} else {
		for_each_cpu(i) {
			sock_release(per_cpu(userspace_crypto_socket, i));
			per_cpu(userspace_crypto_socket, i) = NULL;
		}
	}

	flush_workqueue(wqs);
	destroy_workqueue(wqs);
workqueue_failed:


	return err;
}

int userspace_crypto_netlink(struct nlmsghdr *nlh)
{
	int err = 0;
	int type;
	struct pla_netlink_target *targethdr;

	type = nlh->nlmsg_type;

	switch(type) {
	case NLMSG_PLA_ADD:
		targethdr = NLMSG_DATA(nlh);
		spin_lock_bh(&usercrypto_lock);
		ucrypto_dst = ntohl(targethdr->addr.s_addr);
		spin_unlock_bh(&usercrypto_lock);
		err = 1;
		break;

	default:
		err = 0;
		break;
	}

	return err;
}

void userspace_crypto_uninit(void)
{
	userspace_crypto_init_uninit(0);
	return;
}

int userspace_crypto_init(void)
{
	return userspace_crypto_init_uninit(1);
}


struct pla_crypto_impl userspace_crypto =
{
	.crypto_impl_init = userspace_crypto_init,
	.crypto_impl_uninit = userspace_crypto_uninit,

	.crypto_netlink = userspace_crypto_netlink,

	.call_crypto = userspace_crypto_call
};



