#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>

#include "../src/pla_userspace_crypto.h"
#include "../src/pla_crypto.h"

#include "usercrypto.h"

static int drop_percent = 0;

struct crypto_ops *usercrypto_ops;

int rfd;
int pfd;

/* Mapping between a request and its address */
struct id_entry 
{
	int inuse;                  /* Is this entry in use */
	uint32_t idnum;
	uint64_t seqnum;            /* Gives an 'age' to this entry */
	struct in_addr addr;
};

#define ID_ENTRIES 100

static uint64_t seqnum = 0;
struct id_entry id_entries[ID_ENTRIES];


/* jlu XXX: Needs to be retested for async crypto */
void add_entry(uint32_t idnum, struct in_addr *addr) 
{
	int i;

	int smallest_seq_idx = 0;
	uint64_t smallest_seq = UINT64_MAX;

	for (i=0; id_entries[i].inuse && i < ID_ENTRIES; i++) {
		if (id_entries[i].seqnum < smallest_seq) {
			smallest_seq = id_entries[i].seqnum;
			smallest_seq_idx = i;
		}
	}

	/* All entries are in use. Use the oldest. */
	if (i >= ID_ENTRIES)
		i = smallest_seq_idx;

	id_entries[i].inuse = 1;
	id_entries[i].idnum = idnum;
	id_entries[i].seqnum = seqnum++;
	memcpy(&id_entries[i].addr, addr, sizeof(struct in_addr));

	seqnum++;

	return;
}

/* jlu XXX: Needs to be retested for async crypto */
int find_del_entry(uint32_t idnum, struct in_addr *addr)
{
	int i;

	for (i=0; i < ID_ENTRIES; i++)
		if (id_entries[i].inuse && id_entries[i].idnum == idnum) {
			memcpy(addr, &id_entries[i].addr, sizeof(struct in_addr));
			id_entries[i].inuse = 0;

			return 1;
		}
		
	/* Not found */
	return 0;
}


int open_raw_socket(void)
{
	int fd;
	int zero = 0;
	struct sockaddr_in sin;

	fd = socket(PF_INET, SOCK_RAW, IPPROTO_USERSPACE_CRYPTO);
	if (fd < 0) {
		perror("socket");
		exit(EXIT_FAILURE);
	}

	if (setsockopt(fd, IPPROTO_IP, IP_HDRINCL, &zero, sizeof(zero)) < 0) {
		perror("setsockopt");
		exit(EXIT_FAILURE);
	}

	return fd;
}


int open_packet_socket(void)
{
	int fd;
	int zero = 0;
	struct sockaddr_in sin;

	fd = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_ALL) );
	if (fd < 0) {
		perror("socket");
		exit(EXIT_FAILURE);
	}

	return fd;
}

int usercrypto_result(uint32_t idnum, char *result, int resultlen)
{
	int len;
	char replybuf[1024];

	struct userspace_crypto_reply *reply = (struct userspace_crypto_reply*) replybuf;
	struct sockaddr_in sin;

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = PF_INET;

	if (find_del_entry(idnum, &sin.sin_addr) == 0)
		goto out;   /* Not found */

	memset(reply, 0, sizeof(*reply));
	reply->msg_type = USERSPACE_TYPE_REPLY;
	
	reply->id_num = htonl(idnum);
	reply->task_type = CRYPTO_TASK_TYPE_RESPONSE;
	reply->task_code = CRYPTO_TASK_CODE_OK;

	memcpy(reply->resultdata, result, resultlen);
	reply->resultdata_len = htons(resultlen);

	len = sendto(rfd, reply, sizeof(*reply) + resultlen, 0, (struct sockaddr*) &sin, sizeof(sin));
	if (len < 0) {
		perror("send");
		exit(EXIT_FAILURE);
	} 
	
out:
	return 0;

}

void run() 
{
	int len;
	char inbuf[1500];
	struct sockaddr_in sin;
	
	struct iphdr *iph;
	struct userspace_crypto_req *req;
	
	while(1) {
		socklen_t socklen = 0;
		int id_num;
		
		len = recvfrom(pfd, &inbuf, sizeof(inbuf), 0, NULL, &socklen);
		if (len < 0) {
			perror("recvfrom");
			exit(EXIT_FAILURE);
		} 

		iph = (struct iphdr*) &inbuf;
		if (iph->protocol != IPPROTO_USERSPACE_CRYPTO)
			continue;

		/* assumption: no headers between IPv4 and request */
		req = (struct userspace_crypto_req*) (((char*) iph) + iph->ihl * 4);

		id_num = ntohl(req->id_num);

		if (req->msg_type != USERSPACE_TYPE_REQUEST)
			continue;

		switch(req->task_type) {
		case CRYPTO_TASK_TYPE_VERIFY:
			break;
		case CRYPTO_TASK_TYPE_SIGN:
			break;
		default:
			fprintf(stderr, "Unknown type received.\n");
			break;
		}

		add_entry(id_num, (struct in_addr*) &iph->saddr);
		usercrypto_ops->request(id_num, req->task_type, req->task_data );
	}

	return;
}

int usercrypto_run(struct crypto_ops *ops)
{
	int c;
	int digit_optind = 0;

	if (!ops) {
		fprintf(stderr, "Crypto ops is NULL in usercrypto_run\n");
		exit(EXIT_FAILURE);
	}

	usercrypto_ops = ops;

	rfd = open_raw_socket();
	pfd = open_packet_socket();

	run();

	close(rfd);
	close(pfd);
	
	return EXIT_SUCCESS;
}
